var common = require('./common.js'),
    assert = require('assert'),
    Persistence = require('../core').Persistence,
    Client = require('radar_client').constructor;

exports['given a server'] = {

  before: function(done) {
    common.startRadar(8002, this, done);
  },

  after: function(done) {
    common.endRadar(this, function() {});
    Persistence.disconnect(done);
  },

  beforeEach: function(done) {
    var tasks = 0;
    function next() { tasks++; if (tasks == 2) done(); }
    this.client = new Client()
      .configure({ userId: 123, userType: 0, accountName: 'dev', port: 8002})
      .alloc('test', next);
    Persistence.delWildCard('*:/dev/*', next);
  },

  afterEach: function() {
//    this.client.dealloc('test');
  },

// Presence tests
// - .get(callback)
// - .set('online', ack) / .set('offline', ack)
// - .subscribe(ack)
// - .unsubscribe(ack)
// - .sync(callback)
  'presence: can set("online")': function(done) {
    var client = this.client;
    common.radar().once('set', function(socket, message) {
      assert.equal('set', message.op);
      assert.equal('online', message.value);
      assert.equal(123, message.key);
      client.presence('ticket/21').get(function(message) {
        assert.equal('get', message.op);
        assert.deepEqual(message.value, { 123: 0});
        client.presence('ticket/21').set('offline', function() {
          done();
        });
      });
    });
    this.client.presence('ticket/21').set('online');
  },

  'presence: can set("offline")': function(done) {
    var client = this.client;
    common.radar().once('set', function(socket, message) {
      assert.equal('set', message.op);
      assert.equal('offline', message.value);
      assert.equal(123, message.key);
      // finish executing this request in Radar first
      process.nextTick(function() {
        client.presence('ticket/21').get(function(message) {
          assert.equal('get', message.op);
          assert.deepEqual(message.value, { });
          done();
        });
      });
    });
    this.client.presence('ticket/21').set('offline');
  },

  'presence: can get() using v1 API': function(done) {
    var client = this.client;
    client.presence('ticket/21').get(function(message) {
      assert.equal('get', message.op);
      assert.deepEqual([], message.value);
      client.presence('ticket/21').set('online', function() {
        client.presence('ticket/21').get(function(message) {
          assert.equal('get', message.op);
          assert.deepEqual(message.value, { '123': 0 });
          done();
        });
      });
    });
  },

  'presence: can get() using v2 API': function(done) {
    var client = this.client;
    client.presence('ticket/33').get({ version: 2 }, function(message) {
      assert.equal('get', message.op);
      assert.deepEqual([], message.value);
      client.presence('ticket/33').set('online', function() {
        client.presence('ticket/33').get({ version: 2 }, function(message) {
          assert.equal('get', message.op);
          var expected = {123:{clients:{},userType:0}};
          expected['123'].clients[client._socket.id] = {};
          assert.deepEqual(message.value, expected);
          done();
        });
      });
    });
  },

  'presence: can sync() via v2 API': function(done) {
    // not supported in v1 api because the result.op == "online" which is handled by the message
    // listener but not by the sync() callback
    var client = this.client;

    this.client.presence('ticket/213').set('online', function() {
      client.presence('ticket/213').sync({ version: 2 }, function(message) {
        // sync is implemented as subscribe + get, hence the return op is "get"
        assert.equal('get', message.op);
        var expected = {123:{clients:{},userType:0}};
        expected['123'].clients[client._socket.id] = {};
        assert.deepEqual(message.value, expected);
        done();
      });
    });
  },

// Status tests
// - .set(value, ack)
// - .get(callback)
// - .subscribe(ack)
// - .unsubscribe(ack)
// - .sync(callback)

  'status: can set([String])': function(done) {
    common.radar().once('set', function(client, message) {
      assert.equal('set', message.op);
      assert.equal('foo', message.value);
      assert.equal(123, message.key);
      done();
    });
    this.client.status('voice/status').set('foo');
  },

  'status: can set([Object])': function(done) {
    var client = this.client;
    common.radar().once('set', function(ignore, message) {
      assert.equal('set', message.op);
      assert.deepEqual({ hello: 'world' }, message.value);
      assert.equal(123, message.key);
      client.status('voice/status').get(function(message) {
        assert.equal('get', message.op);
        assert.deepEqual( { hello: 'world' }, message.value['123']);
        done();
      });
    });
    this.client.status('voice/status').set({ hello: 'world' });
  },

  'status: can get([String])': function(done) {
    var client = this.client;

    this.client.status('voice/status').set('foo');

    client.status('voice/status').get(function(message) {
      assert.equal('get', message.op);
      assert.deepEqual({ 123: 'foo'}, message.value);
      client.status('voice/status').set('bar');
      client.status('voice/status').get(function(message) {
        assert.equal('get', message.op);
        assert.deepEqual({ 123: 'bar'}, message.value);
        done();
      });
    });
  },

  'status: can get([Object])': function(done) {
    var client = this.client;

    this.client.status('voice/status').set({'foo': 'bar'});

    client.status('voice/status').get(function(message) {
      assert.equal('get', message.op);
      assert.deepEqual({ 123: {'foo': 'bar'}}, message.value);
      client.status('voice/status').set({'foo2': 'bar2'});
      client.status('voice/status').get(function(message) {
        assert.equal('get', message.op);
        assert.deepEqual({ 123: {'foo2': 'bar2'}}, message.value);
        done();
      });
    });
  },

  'status: can subscribe()': function(done) {
    this.timeout(10000);

    var client = this.client;
    var client2 = new Client()
      .configure({ userId: 234, userType: 0, accountName: 'dev', port: 8002})
      .alloc('test', function() {

        client2.status('voice/status')
          .once(function(message) {
            assert.equal('set', message.op);
            assert.equal('foo', message.value);
            assert.equal('123', message.key);
            done();
          })
          .subscribe(function() {
            client.status('voice/status').set('foo');
          });
      });
  },


  'status: can sync()': function(done) {
    var client = this.client;

    this.client.status('voice/status').set('foo', function() {
      client.status('voice/status').sync(function(message) {
        // sync is implemented as subscribe + get, hence the return op is "get"
        assert.equal('get', message.op);
        assert.deepEqual({ 123: 'foo'}, message.value);
        done();
      });
    });
  },

// Message list tests
// - .subscribe('channel')
// - .unsubscribe('channel')
// - .sync('channel')
// - .publish('channel', message)

  'message: can subscribe()': function(done) {
    var client = this.client;
    // test.expect(2);
    common.radar().when('subscribe', function(client, msg) {
      var match = (msg.op == 'subscribe' && msg.to == 'message:/dev/test');
      if (match) {
        assert.equal('subscribe', msg.op);
        assert.equal('message:/dev/test', msg.to);
        done();
      }
      return match;
    });
    // optional (due to new "connect-if-configured-logic"): client.alloc('test');
    client.message('test').subscribe();
  },

  'message: can publish([Object])': function(done) {
    var client = this.client;
    // test.expect(2);
    var message = { state: 'other'};

    client.message('test').when(function(msg) {
      if(msg.value && msg.value.state && msg.value.state == 'other') {
        assert.equal('message:/dev/test', msg.to);
        assert.equal('other', msg.value.state);
        done();
        return true;
      }
      return false;
    });
    client.message('test').subscribe().publish(message);
  },

  'message: can publish([String])': function(done) {
    var client = this.client;
    // test.expect(2);
    var message = '{ "state": "other"}';

    client.message('test').when(function(msg) {

      assert.equal('message:/dev/test', msg.to);

      assert.equal('string', typeof msg.value);
      assert.equal('{ "state": "other"}', msg.value);

      done();

    });
    client.message('test').subscribe().publish(message);
  },

  'message: can sync([String])': function(done) {
    var client = this.client,
        message = 'foobar',
        assertions = 0;
    Persistence.persistOrdered('message:/dev/test', message, function() {
      client.message('test').on(function(msg) {
        assert.equal('foobar', msg);
        assertions++;
      }).sync('test');
      setTimeout(function() {
        assert.equal(1, assertions);
        done();
      }, 50);
    });
  },

  'message: can sync([Object])': function(done) {
    var client = this.client,
        message = { foo: 'bar' },
        assertions = 0;
    Persistence.persistOrdered('message:/dev/test', message, function() {
      client.message('test').on(function(msg) {
        assert.equal('bar', msg.foo);
        assertions++;
      }).sync('test');
      setTimeout(function() {
        assert.equal(1, assertions);
        done();
      }, 50);
    });
  }
};